# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
import json
import uuid
from typing import Any, Dict, List

from django.conf import settings
from django.db import models
from django.http import HttpRequest
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from jsonfield import JSONField
from translated_fields import TranslatedField

from blue_krill.models.fields import EncryptField

# Base Models start


class AuditedModel(models.Model):
    """Audited model with 'created' and 'updated' fields."""

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta(object):
        abstract = True


class UuidAuditedModel(AuditedModel):
    """Add a UUID primary key to an :class:`AuditedModel`."""

    uuid = models.UUIDField(
        'UUID', default=uuid.uuid4, primary_key=True, editable=False, auto_created=True, unique=True
    )

    class Meta(object):
        abstract = True


# Base Models end


class SpecDefinition(UuidAuditedModel):
    """Specification definition"""

    index = models.IntegerField(verbose_name=_('顺序'), default=0)
    name = models.CharField(verbose_name=_('名称'), unique=True, max_length=64, null=True)
    display_name = TranslatedField(models.CharField(verbose_name=_('展示名称'), max_length=128, blank=True))
    description = models.TextField(verbose_name=_('描述'), blank=True)
    recommended_value = models.CharField(verbose_name=_('推荐值'), max_length=64, blank=True, null=True)

    class Meta:
        verbose_name_plural = verbose_name = _('规格定义')
        ordering = ['index', 'name']

    def __str__(self):
        return f"{self.display_name}"

    def save(self, *args, **kwargs):
        if not self.display_name:
            self.display_name = self.name
        super().save(*args, **kwargs)


class Specification(UuidAuditedModel):
    """Specification value"""

    definition = models.ForeignKey(SpecDefinition, verbose_name=_('定义'), on_delete=models.CASCADE)
    value = models.CharField(verbose_name=_('值'), max_length=64)
    display_name = TranslatedField(models.CharField(verbose_name=_('展示名称'), max_length=64, blank=True))

    class Meta:
        verbose_name_plural = verbose_name = _('规格')
        unique_together = ['definition', 'value']

    def __str__(self):
        return f"{self.definition}:{self.display_name}"

    def save(self, *args, **kwargs):
        if not self.display_name:
            self.display_name = self.value
        super().save(*args, **kwargs)


class Service(UuidAuditedModel):
    """Service model for PaaS"""

    name = models.CharField(verbose_name='服务名称', unique=True, max_length=64)
    category = models.IntegerField(verbose_name='服务分类')
    display_name = TranslatedField(models.CharField(verbose_name=_('服务全称'), max_length=128))
    logo_url = models.URLField(verbose_name='服务 logo 地址', null=True, blank=True)
    logo = models.TextField(verbose_name='服务 logo base64', null=True, blank=True)

    description = TranslatedField(models.CharField(verbose_name=_('简介'), max_length=1024, blank=True))
    long_description = TranslatedField(models.TextField(verbose_name=_('详细介绍'), blank=True))
    instance_tutorial = TranslatedField(models.TextField(verbose_name=_('实例内页介绍'), blank=True))

    available_languages = models.CharField(verbose_name=_('支持编程语言'), max_length=1024, null=True, blank=True)
    config = JSONField(default=dict, blank=True)
    is_active = models.BooleanField(verbose_name='是否可用', default=True)
    is_visible = models.BooleanField(verbose_name="是否可见", default=True)

    specifications = models.ManyToManyField(SpecDefinition, verbose_name='规格定义', blank=True)

    class Meta:
        verbose_name_plural = verbose_name = '服务'

    def __str__(self):
        return f"{self.name}"


class ServiceInstance(UuidAuditedModel):
    """specific info of service"""

    _prepared_fields_data = None

    service = models.ForeignKey('Service', verbose_name='服务', null=True, on_delete=models.SET_NULL)
    plan = models.ForeignKey(
        'Plan', verbose_name='方案', null=True, help_text="当前仅当迁移的增强服务实例, 会没有 plan", on_delete=models.SET_NULL
    )
    config = JSONField(default=dict, blank=True)
    credentials = EncryptField(default="")
    to_be_deleted = models.BooleanField(default=False)

    class Meta:
        verbose_name_plural = verbose_name = '服务实例'

    def __str__(self):
        return f"{self.service}-{self.plan}-{self.uuid}"

    def get_credentials(self):
        return json.loads(self.credentials)

    @property
    def from_client_side(self):
        """描述该实例是否由 'client-side' 创建, 远程增强服务中负责存储实例的 credentials, config 信息"""
        return self.plan is None

    def prerender(self, request: HttpRequest):
        """Prepare for rendering data to client"""
        # Allow other services to override default render logic
        func = import_string(
            getattr(settings, 'PAAS_SERVICE_SVC_INSTANCE_RENDER_FUNC', 'paas_service.models.render_instance_data')
        )
        data = func(request, self)
        self._prepared_fields_data = data

    def render(self):
        if not self._prepared_fields_data:
            raise ValueError('no prerendered data found, please call prerender method first')
        return self._prepared_fields_data


class ServiceInstanceConfig(UuidAuditedModel):
    """Extra config for instance"""

    instance = models.OneToOneField(ServiceInstance, verbose_name='服务实例', on_delete=models.CASCADE)
    paas_app_info = JSONField(default=dict, blank=True, verbose_name='平台应用信息')

    def was_initialized(self) -> bool:
        """Return if current config object was initialized"""
        return bool(self.paas_app_info)


class InstanceDataRepresenter:
    """represents service instance"""

    def __init__(self, request: HttpRequest, instance: ServiceInstance):
        self.request = request
        self.instance = instance

        # Set initial values for config/credentials
        self._config = self.instance.config
        self._credentials = json.loads(self.instance.credentials)
        # Meta config is a field which store meta options
        self.meta_config: Dict[str, Any] = {}

    def set_hidden_fields(self, hidden_fields: List):
        """Set fields as hidden"""
        self.meta_config['should_hidden_fields'] = hidden_fields

    def set_removed_fields(self, fields: List):
        self.meta_config['should_remove_fields'] = fields

    def represent(self) -> Dict:
        return {
            'config': {'__meta__': self.meta_config, **self._config},
            'credentials': self._credentials,
        }


def render_instance_data(request: HttpRequest, instance: ServiceInstance) -> Dict:
    """Default render function for rendering service instance data"""
    return InstanceDataRepresenter(request, instance).represent()


class Plan(UuidAuditedModel):
    name = models.CharField(verbose_name='方案名称', max_length=64)

    # the "properties" field stores custom properties of a plan object. the property itself is
    # nothing more than a simple annotation. But the service hub which it was registed to may
    # take special actions according to these annotations.
    #
    # Some example properties: {"region": "r1", "as_env_default": "prod" ... ...}
    properties = JSONField(default=dict, blank=True)

    description = models.CharField(verbose_name='方案简介', max_length=1024, blank=True)
    config = EncryptField(verbose_name='方案配置', default="", blank=True)
    is_active = models.BooleanField(verbose_name='是否可用', default=True)
    service = models.ForeignKey('Service', related_name='plans', verbose_name='服务', on_delete=models.CASCADE)

    specifications = models.ManyToManyField(Specification, verbose_name='规格', blank=True)

    class Meta:
        verbose_name_plural = verbose_name = '方案'
        unique_together = ("service", "name")

    def __str__(self):
        return f"{self.name}-{self.service.name}"

    def get_config(self):
        return json.loads(self.config)

    @cached_property
    def full_specifications(self) -> Dict[str, Any]:
        # 已定义的全量规格,默认为 None
        specifications = dict.fromkeys(d.name for d in self.service.specifications.all())

        # 填充真实值
        for spec in self.specifications.filter(definition__name__in=specifications.keys(),).prefetch_related(
            "definition"
        ):  # type: Specification
            specifications[spec.definition.name] = spec.value

        return specifications


class ResourceId(models.Model):
    """utility model for avoiding name conflict"""

    namespace = models.CharField(max_length=32)
    uid = models.CharField(max_length=64, null=False, unique=True, db_index=True)

    class Meta(object):
        unique_together = ('namespace', 'uid')

    def __str__(self):
        return f"{self.namespace}-{self.uid}"



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
# Generated by Django 1.11.20 on 2020-01-03 04:09
from __future__ import unicode_literals

import uuid

import django.db.models.deletion
import jsonfield.fields
from django.db import migrations, models

import blue_krill.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('paas_service', '0002_serviceinstance_need_to_delete'),
    ]

    operations = [
        migrations.CreateModel(
            name='SpecDefinition',
            fields=[
                (
                    'uuid',
                    models.UUIDField(
                        auto_created=True,
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name='UUID',
                    ),
                ),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('index', models.IntegerField(default=0, verbose_name='顺序')),
                ('name', models.CharField(max_length=64, null=True, unique=True, verbose_name='名称')),
                ('display_name', models.CharField(blank=True, max_length=128, verbose_name='展示名称')),
                ('description', models.TextField(blank=True, verbose_name='描述')),
                ('recommended_value', models.CharField(blank=True, max_length=64, null=True, verbose_name='推荐值')),
            ],
            options={
                'verbose_name': '规格定义',
                'verbose_name_plural': '规格定义',
                'ordering': ['index', 'name'],
            },
        ),
        migrations.CreateModel(
            name='Specification',
            fields=[
                (
                    'uuid',
                    models.UUIDField(
                        auto_created=True,
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name='UUID',
                    ),
                ),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('value', models.CharField(max_length=64, verbose_name='值')),
                ('display_name', models.CharField(blank=True, max_length=64, verbose_name='展示名称')),
                (
                    'definition',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='paas_service.SpecDefinition',
                        verbose_name='定义',
                    ),
                ),
            ],
            options={
                'verbose_name': '规格',
                'verbose_name_plural': '规格',
            },
        ),
        migrations.AlterModelOptions(
            name='plan',
            options={'verbose_name': '方案', 'verbose_name_plural': '方案'},
        ),
        migrations.AlterModelOptions(
            name='service',
            options={'verbose_name': '服务', 'verbose_name_plural': '服务'},
        ),
        migrations.AlterModelOptions(
            name='serviceinstance',
            options={'verbose_name': '服务实例', 'verbose_name_plural': '服务实例'},
        ),
        migrations.AlterField(
            model_name='plan',
            name='config',
            field=blue_krill.models.fields.EncryptField(blank=True, default='', verbose_name='方案配置'),
        ),
        migrations.AlterField(
            model_name='plan',
            name='properties',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='plan',
            name='service',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='plans',
                to='paas_service.Service',
                verbose_name='服务',
            ),
        ),
        migrations.AlterField(
            model_name='service',
            name='config',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='serviceinstance',
            name='config',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='serviceinstance',
            name='plan',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to='paas_service.Plan', verbose_name='方案'
            ),
        ),
        migrations.AlterField(
            model_name='serviceinstance',
            name='service',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to='paas_service.Service', verbose_name='服务'
            ),
        ),
        migrations.AddField(
            model_name='plan',
            name='specifications',
            field=models.ManyToManyField(blank=True, to='paas_service.Specification', verbose_name='规格'),
        ),
        migrations.AddField(
            model_name='service',
            name='specifications',
            field=models.ManyToManyField(blank=True, to='paas_service.SpecDefinition', verbose_name='规格定义'),
        ),
        migrations.AlterUniqueTogether(
            name='specification',
            unique_together=set([('definition', 'value')]),
        ),
    ]



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
import logging
from typing import Optional

from .backends import AuthFailedError, JWTClientAuthenticater

logger = logging.getLogger(__name__)


class VerifiedClientMiddleware(object):
    """This middleware will inject request.verifed_client if current request has a valid
    signed JWT token in request headers
    """

    def __init__(self, get_response):
        self.get_response = get_response

    @staticmethod
    def get_token(request) -> Optional[str]:
        """Get token from request, return None if no token can be found"""
        auth_header_val = request.META.get('HTTP_AUTHORIZATION')
        if not auth_header_val:
            return None

        try:
            prefix, token = auth_header_val.split(None, 1)
        except ValueError:
            logger.warning('Invalid authorization header value')
            return None

        return token

    def __call__(self, request):
        token = self.get_token(request)
        request.client = None
        if token:
            try:
                ret = JWTClientAuthenticater().authenticate(token=token)
                request.client = ret.client
            except AuthFailedError:
                pass

        response = self.get_response(request)
        return response



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
import json
import logging
from contextlib import contextmanager

from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404, redirect, reverse
from django.utils.decorators import method_decorator
from django.views import View
from paas_service import __version__, serializers
from paas_service.auth.backends import InstanceAuthBackend, InstanceAuthFailed
from paas_service.auth.decorator import verified_client_required, verified_client_role_require
from paas_service.base_vendor import ArgumentInvalidError, InstanceData, OperationFailed, get_provider_cls
from paas_service.mixins import LoginRequiredMixin
from paas_service.models import Plan, Service, ServiceInstance, ServiceInstanceConfig
from paas_service.utils import parse_redirect_params
from rest_framework import status, viewsets
from rest_framework.response import Response

logger = logging.getLogger(__name__)


m_verified_client_required = method_decorator(verified_client_required)


class AuthenticationViewSet(LoginRequiredMixin, View):
    """View for authentication"""

    def get(self, request):
        """Accepts and validates a JWT token parameter in GET string, will redirect to the index
        page of the instance after a successfully validation
        """
        try:
            instance = InstanceAuthBackend().invoke(request)
        except InstanceAuthFailed as e:
            logger.error(f'unable to authenticate current request: {e}')
            raise Http404

        view_name, params = parse_redirect_params(
            request.GET.get("redirect_url", None), instance_id=str(instance.uuid)
        )
        return redirect(reverse(view_name, kwargs=params))


class SvcBasicViewSet(viewsets.ViewSet):
    """Basic views for services"""

    def get_meta_info(self, request):
        """Return service meta info"""
        return Response({'version': __version__})


class ServiceManageViewSet(viewsets.ViewSet):
    """Manage view for services"""

    @verified_client_role_require('internal_platform')
    def list(self, request):
        services = Service.objects.filter(is_active=True).prefetch_related('plans')
        slz = serializers.ServiceListSLZ(services, many=True)
        return Response(slz.data)

    @verified_client_role_require('internal_platform')
    def update(self, request, service_id):
        service = get_object_or_404(Service, pk=service_id)
        slz = serializers.ServiceUpsertSLZ(data=request.data, instance=service, partial=True)
        slz.is_valid(True)
        slz.save()
        return Response(status=status.HTTP_204_NO_CONTENT)

    @verified_client_role_require('internal_platform')
    def create(self, request):
        slz = serializers.ServiceUpsertSLZ(data=request.data)
        slz.is_valid(True)
        slz.save()
        return Response(status=status.HTTP_201_CREATED)

    @verified_client_role_require('internal_platform')
    def destroy(self, request, service_id):
        service = get_object_or_404(Service, pk=service_id)
        service.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


class PlanManageViewSet(viewsets.ViewSet):
    """Manage view for Plan"""

    @verified_client_role_require('internal_platform')
    def create(self, request):
        slz = serializers.PlanUpsertSLZ(data=request.data)
        slz.is_valid(True)
        slz.save()
        return Response(status=status.HTTP_201_CREATED)

    @verified_client_role_require('internal_platform')
    def update(self, request, plan_id):
        plan = get_object_or_404(Plan, pk=plan_id)
        slz = serializers.PlanUpsertSLZ(data=request.data, instance=plan, partial=True)
        slz.is_valid(True)
        plan = slz.save()
        return Response(status=status.HTTP_204_NO_CONTENT)

    @verified_client_role_require('internal_platform')
    def destroy(self, request, plan_id):
        plan = get_object_or_404(Plan, pk=plan_id)
        plan.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


class SvcInstanceViewSet(viewsets.ViewSet):
    """Instance view set"""

    @verified_client_role_require('internal_platform')
    def provision(self, request, service_id, instance_id):
        """Provision a new instance

        an example request data:

            {
                "plan_id": "f8ad12f2-4dd5-4871-8e31-9a1f9f3795a2",
                "params": {
                    "username": "example-bucket"
                }
            }

        """
        if ServiceInstance.objects.filter(pk=instance_id):
            # TODO: Replace with standard error code
            return Response(data={'detail': f'instance {instance_id} already exists'}, status=400)

        service = get_object_or_404(Service, pk=service_id)
        plan_id = request.data.get('plan_id')
        if not plan_id:
            return Response(data={'detail': 'plan_id is missing'}, status=400)

        provider_cls = get_provider_cls()
        plan = get_object_or_404(Plan, service=service, pk=plan_id)

        plan_config = json.loads(plan.config)
        params = request.data.get('params', {})
        with wrap_provider_action_exc('create instance') as ret:
            instance_data = provider_cls(**plan_config).create(params=params)

        if ret.has_error:
            return ret.response

        service_instance = ServiceInstance.objects.create(
            pk=instance_id,
            service=service,
            plan=plan,
            config=instance_data.config,
            credentials=json.dumps(instance_data.credentials),
        )
        service_instance.prerender(request)
        return Response(serializers.ServiceInstanceSLZ(service_instance).data, status=201)

    @m_verified_client_required
    def retrieve(self, request, instance_id):
        """Retrieve an instance"""
        instance = get_object_or_404(ServiceInstance, pk=instance_id, to_be_deleted=False)
        instance.prerender(request)
        return Response(serializers.ServiceInstanceSLZ(instance).data)

    @m_verified_client_required
    def retrieve_by_fields(self, request, service_id):
        """Retrieve an instance by fields"""
        qs = ServiceInstance.objects.filter(service__uuid=service_id)
        # 由于 name 在 credentials 加密字段中， 所以不能直接使用 filter 过滤
        # 只能遍历所有 ServiceInstance 来查找
        name = request.GET.get("name")
        for instance in qs:
            credentials = instance.get_credentials()
            if credentials.get('name') == name:
                instance.prerender(request)
                return Response(serializers.ServiceInstanceSLZ(instance).data)
        return Response(status=404)

    @verified_client_role_require('internal_platform')
    def update(self, request, instance_id):
        return Response(data={'detail': 'not available'}, status=400)

    @verified_client_role_require('internal_platform')
    def destroy(self, request, instance_id):
        """Destroy an instance"""
        instance = get_object_or_404(ServiceInstance, pk=instance_id)
        provider_cls = get_provider_cls()
        instance_data = InstanceData(
            credentials=json.loads(instance.credentials),
            config=instance.config,
        )
        plan_config = json.loads(instance.plan.config)

        with wrap_provider_action_exc(f'remove instance {instance.uuid}') as ret:
            provider_cls(**plan_config).delete(instance_data)

        if ret.has_error:
            return ret.response

        instance.delete()
        return Response(status=204)

    @verified_client_role_require('internal_platform')
    def async_destroy(self, request, instance_id):
        """Destroy an instance async, record it into db"""
        instance = get_object_or_404(ServiceInstance, pk=instance_id)

        try:
            instance.to_be_deleted = True
            instance.save(update_fields=["to_be_deleted"])
        except Exception as e:
            logger.exception("mark instance need to delete failed")
            return Response(data={'detail': f'unable to mark instance<{instance_id}> to delete: {e}'})

        return Response(status=204)


class ClientSideSvcInstanceViewSet(viewsets.ViewSet):
    @verified_client_role_require('internal_platform')
    def create(self, request, service_id, instance_id):
        """create service instance with the existing config and credentials.

        an example request data:
            {
                "config": {
                    ...
                },
                "credentials": {
                    ...
                }
            }

        """
        if ServiceInstance.objects.filter(pk=instance_id):
            # TODO: Replace with standard error code
            return Response(data={'detail': f'instance {instance_id} already exists'}, status=400)

        slz = serializers.ServiceInstanceBinderSLZ(data=request.data)
        slz.is_valid(raise_exception=True)
        data = slz.validated_data

        config = data["config"]
        credentials = data["credentials"]
        service = get_object_or_404(Service, pk=service_id)
        service_instance = ServiceInstance.objects.create(
            pk=instance_id,
            service=service,
            plan=None,
            config=config,
            credentials=json.dumps(credentials),
        )
        service_instance.prerender(request)
        return Response(serializers.ServiceInstanceSLZ(service_instance).data, status=201)

    @verified_client_role_require('internal_platform')
    def destroy(self, request, instance_id):
        """Unbind the instance without reclaiming resources."""
        instance = get_object_or_404(ServiceInstance, pk=instance_id)
        if not instance.from_client_side:
            return Response(data={'detail': f'instance {instance_id} is not a client-side instance'}, status=400)
        instance.delete()
        return Response(status=204)


class SvcInstanceConfigViewSet(viewsets.ViewSet):
    """InstanceConfig view"""

    @verified_client_role_require('internal_platform')
    def retrieve(self, request, instance_id):
        """Retrieve an instance's config"""
        instance = get_object_or_404(ServiceInstance, pk=instance_id)
        config, _ = ServiceInstanceConfig.objects.get_or_create(instance=instance)
        if not config.was_initialized():
            return Response(data={'detail': 'config was not initialized yet'}, status=400)

        slz = serializers.InstanceConfigSLZ(config)
        return Response(slz.data)

    @verified_client_role_require('internal_platform')
    def update(self, request, instance_id):
        """Update an instance's config"""
        instance = get_object_or_404(ServiceInstance, pk=instance_id)
        slz = serializers.InstanceConfigSLZ(data=request.data)
        slz.is_valid(raise_exception=True)

        # Update config
        config, _ = ServiceInstanceConfig.objects.update_or_create(
            instance=instance, defaults={'paas_app_info': slz.validated_data['paas_app_info']}
        )
        resp_slz = serializers.InstanceConfigSLZ(config)
        return Response(resp_slz.data)


@contextmanager
def wrap_provider_action_exc(action_name: str):
    """use this as a context manager to transform provider operation exceptions into user
    friendly error messages
    """

    class OpResult:
        def __init__(self):
            self.response = None
            self.has_error = False

        def set_ok(self):
            self.has_error = False

        def set_fail_with_resp(self, resp: HttpResponse):
            self.has_error = True
            self.response = resp

    ret = OpResult()
    try:
        yield ret
    except ArgumentInvalidError as e:
        logger.exception(f'argument error when {action_name}')
        ret.set_fail_with_resp(Response(data={'detail': f'params error: {e}'}, status=400))
    except OperationFailed as e:
        logger.exception(f'operation failed while {action_name}')
        ret.set_fail_with_resp(Response(data={'detail': f'operation failed: {e}'}, status=500))
    except Exception:
        logger.exception(f'error while {action_name}')
        ret.set_fail_with_resp(Response(data={'detail': f'unable to {action_name}'}, status=500))
    else:
        ret.set_ok()



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
from typing import Optional

from paas_service.models import Plan, Service, ServiceInstance, SpecDefinition, Specification
from rest_framework import serializers


class SpecDefinitionSLZ(serializers.ModelSerializer):
    class Meta(object):
        model = SpecDefinition
        exclude = ['index', 'uuid', 'created', 'updated']


class ServiceSLZ(serializers.ModelSerializer):
    specifications = serializers.ListField(child=SpecDefinitionSLZ(), source='specifications.all')
    logo = serializers.SerializerMethodField()
    config = serializers.JSONField()

    def get_logo(self, obj):
        # type: (Service) -> Optional[str]
        if obj.logo:
            return obj.logo
        elif obj.logo_url:
            return obj.logo_url
        else:
            return None

    class Meta(object):
        model = Service
        exclude = ['created', 'updated']


class PlanSLZ(serializers.ModelSerializer):
    properties = serializers.JSONField()
    config = serializers.JSONField(source="get_config")
    specifications = serializers.DictField(source='full_specifications')

    class Meta(object):
        model = Plan
        exclude = ['created', 'updated']


class ServiceListSLZ(ServiceSLZ):
    plans = serializers.ListField(child=PlanSLZ(), source='plans.all')


class ServiceInstanceSLZ(serializers.ModelSerializer):
    def to_representation(self, instance: ServiceInstance):
        instance_info = super(ServiceInstanceSLZ, self).to_representation(instance)
        instance_info.update(instance.render())
        return instance_info

    class Meta(object):
        model = ServiceInstance
        exclude = ['config', 'credentials']


class ServiceInstanceBinderSLZ(serializers.Serializer):
    config = serializers.DictField(required=False, default=dict)
    credentials = serializers.DictField(required=True)


class PaaSAppInfoSLZ(serializers.Serializer):
    """Serializer for Application Info from Blueking PaaS.

    Only `app_id` was required.
    """

    app_id = serializers.CharField(max_length=64)
    app_code = serializers.CharField(max_length=64, default='')
    app_name = serializers.CharField(max_length=64, default='')
    module = serializers.CharField(max_length=64, default='')
    environment = serializers.CharField(max_length=32, default='unknown')


class InstanceConfigSLZ(serializers.Serializer):
    paas_app_info = PaaSAppInfoSLZ(required=True)


############################
# Serializer for admin api #
############################
class _SpecDefinitionUpsertSLZ(serializers.ModelSerializer):
    name = serializers.CharField(help_text='名称', max_length=64)

    def to_internal_value(self, data):
        try:
            instance = SpecDefinition.objects.get(name=data["name"])
            for k, v in data.items():
                setattr(instance, k, v)
        except SpecDefinition.DoesNotExist:
            instance = SpecDefinition(**data)
        return instance

    class Meta(object):
        model = SpecDefinition
        exclude = ['index', 'uuid', 'created', 'updated']


class ServiceUpsertSLZ(serializers.ModelSerializer):
    """Usage
    # For Update
    >>> slz = ServiceUpsertSLZ(data={}, instance=service)
    >>> slz.is_valid(True)
    >>> slz.save()

    # For Create
    >>> slz = ServiceUpsertSLZ(data={})
    >>> slz.is_valid(True)
    >>> slz.save()
    """

    specifications = serializers.ManyRelatedField(child_relation=_SpecDefinitionUpsertSLZ())
    config = serializers.JSONField()

    def update(self, instance, validated_data):
        for spec_def in validated_data["specifications"]:
            spec_def.save()
        return super().update(instance, validated_data)

    def create(self, validated_data):
        for spec_def in validated_data["specifications"]:
            spec_def.save()
        return super().create(validated_data)

    class Meta(object):
        model = Service
        fields = "__all__"


class _SpecificationUpsertSLZ(serializers.JSONField):
    def to_internal_value(self, data):
        specifications = []
        for key, value in data.items():
            definition = SpecDefinition.objects.get(name=key)
            try:
                specifications.append(Specification.objects.get(definition=definition, value=value))
            except Specification.DoesNotExist:
                specifications.append(Specification.objects.create(definition=definition, value=value))
        return specifications


class PlanUpsertSLZ(serializers.ModelSerializer):
    """Usage
    # For Update
    >>> slz = PlanUpsertSLZ(data={}, instance=plan)
    >>> slz.is_valid(True)
    >>> slz.save()

    # For Create
    >>> slz = PlanUpsertSLZ(data={})
    >>> slz.is_valid(True)
    >>> slz.save()
    """

    def save(self, **kwargs):
        # 剔除 region 字段, 避免 create 出错
        region = self.validated_data.pop("region", None)
        instance = super().save(**kwargs)
        # 更新 region 字段.
        if region:
            instance.properties["region"] = region
            instance.save()
        return instance

    region = serializers.CharField()

    properties = serializers.JSONField()
    specifications = _SpecificationUpsertSLZ()

    class Meta(object):
        model = Plan
        exclude = ['created', 'updated']



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
import logging
from collections.abc import Mapping
from functools import wraps

from django.core.exceptions import PermissionDenied
from django.http import HttpResponseServerError
from django.utils.decorators import method_decorator

logger = logging.getLogger(__name__)


def instance_authorized_require(func):
    """Decorator for view methods which use an "instance_id" as first argument,
    will check session to see if instance_id requested has been authorized via
    backends.
    """

    @wraps(func)
    def wrapper(request, instance_id, *args, **kwargs):
        authorized_instances = request.session.get('authorized_instances', {})
        if not isinstance(authorized_instances, Mapping):
            logger.error('authorized_instances type error')
            return HttpResponseServerError('internal server error, please try again')

        # Check if instance has been authorized
        if instance_id not in authorized_instances:
            raise PermissionDenied()

        return func(request, instance_id, *args, **kwargs)

    return wrapper


def verified_client_required(func):
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        client = getattr(request, 'client', None)
        if client and client.is_verified():
            return func(request, *args, **kwargs)
        else:
            raise PermissionDenied()

    return wrapper


def verified_client_role_require(role_name):
    """check if role exist with covering features of the above decorator"""

    @method_decorator
    def _wrapper(func):
        @wraps(func)
        def _wrapped(request, *args, **kwargs):
            client = getattr(request, 'client', None)
            if client and client.is_verified() and role_name == client.role:
                return func(request, *args, **kwargs)
            else:
                raise PermissionDenied()

        return _wrapped

    return _wrapper



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
import logging
import time
from typing import Dict, Tuple

import jwt
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.handlers.wsgi import WSGIRequest
from jwt.exceptions import DecodeError
from paas_service.models import ServiceInstance

logger = logging.getLogger(__name__)


class AuthFailedError(Exception):
    """Raised when one authenticater fails"""


class InstanceAuthFailed(Exception):
    pass


class Client:
    def __init__(self, name, auth_backend_type=None, role=None):
        self.name = name
        self.auth_backend_type = auth_backend_type
        self.role = role

    def __str__(self):
        return self.name

    def is_verified(self):
        return True

    @classmethod
    def from_jwt_settings(cls, data):
        for key in ('iss', 'algorithm', 'key'):
            if not data.get(key):
                raise ValueError(f'error: {key} is required')
        return cls(name=data['iss'], auth_backend_type='jwt')


class AuthResult:
    """result for authentication"""

    def __init__(self, client: Client, extra_payload: Dict):
        self.client = client
        self.extra_payload = extra_payload


class JWTClientAuthenticater:
    """Authenticater using JWT clients"""

    def __init__(self):
        try:
            self.jwt_clients = settings.PAAS_SERVICE_JWT_CLIENTS
        except AttributeError:
            raise ImproperlyConfigured("PAAS_SERVICE_JWT_CLIENTS is not configured")

    def authenticate(self, token: str) -> AuthResult:
        """Authenticate a given token

        :raises: AuthFailedError
        """
        try:
            payload, client = self.parse(token)
        except ValueError as e:
            logger.warning(f'authentication failed, token is invalid: {e}')
            raise AuthFailedError('token is invalid')
        else:
            logger.info(f'successfully authenticated token issued by {client}')
            return AuthResult(client=client, extra_payload=payload)

    def parse(self, token: str) -> Tuple[Dict, Client]:
        """Parse given token

        :raises: ValueError when token is invalid
        """
        for client in self.jwt_clients:
            try:
                payload = jwt.decode(token, client['key'], algorithms=[client['algorithm']])
            except DecodeError:
                logger.debug(f'Unable to decode token using {client["iss"]}\'s secret')
                continue

            if not self._validate_payload(client, payload):
                logger.debug('payload validation failed')
                continue

            client_ins = Client.from_jwt_settings(client)
            client_ins.role = payload.get('role')
            return payload, client_ins
        raise ValueError('token is not a valid JWT token')

    @staticmethod
    def _validate_payload(client: Dict, payload: Dict) -> bool:
        """Validates given JWT payload, a valid payload must contains at least 2 fields:
        'iss' and 'expires_at'.
        """
        expires_at = payload.get('expires_at')
        try:
            expires_at = int(expires_at)  # type: ignore
        except (TypeError, ValueError):
            logger.warning("token's expires_at is empty or invalid")
            return False
        if expires_at < time.time():
            logger.warning("token has already expired")
            return False

        iss = payload.get('iss')
        if not (iss and iss == client['iss']):
            logger.warning("issuer name in payload does not match client's config")
            return False
        return True


class InstanceAuthBackend:
    TOKEN_KEY = 'token'

    def __init__(self):
        self.authenticaters = [JWTClientAuthenticater()]

    def get_token(self, request: WSGIRequest):
        return request.GET.get(self.TOKEN_KEY, None)

    def invoke(self, request: WSGIRequest) -> ServiceInstance:
        """Invoke currrent bakend

        :raises: InstanceAuthFailed
        """
        token = self.get_token(request)
        if not token:
            raise InstanceAuthFailed('token parameter is missing')

        for auth in self.authenticaters:
            try:
                result = auth.authenticate(token)
                break
            except AuthFailedError:
                continue
        else:
            raise InstanceAuthFailed('token is invalid, can not be authenticated')

        instance_id = result.extra_payload.get('service_instance_id')
        if not instance_id:
            logger.warning('service_instance_id field is missing in token payload')
            raise InstanceAuthFailed('no service_instance_id field')

        try:
            instance = ServiceInstance.objects.get(pk=str(instance_id))
        except ServiceInstance.DoesNotExist:
            logger.warning(f'can not find a instance via id: {instance_id}')
            raise InstanceAuthFailed('service_instance_id is invalid')

        # Update session
        request.session['authorized_instances'] = {
            str(instance.uuid): True,
            **request.session.get('authorized_instances', {}),
        }
        return instance


def sign_instance_token(client_name: str, instance_uuid: str) -> str:
    """A quick access to sign a JWT token which is related to instance"""
    return _sign_token(client_name=client_name, additional_payload={"service_instance_id": instance_uuid})


def sign_role_aware_token(client_name: str, role_name: str) -> str:
    """A quick access to sign a role-aware JWT token"""
    return _sign_token(client_name=client_name, additional_payload={"role": role_name})


def _sign_token(client_name: str, additional_payload: Dict) -> str:
    """Sign a new JWT token using given client's crendentials, this token can be used
    for an InstanceAuthBackend's authentication process.

    :returns: a JWT token which will expires after 3600 seconds
    :raises: ValueError if given client name is not valid
    """
    for client_data in settings.PAAS_SERVICE_JWT_CLIENTS:
        if client_name == client_data['iss']:
            payload = {'iss': client_name, 'expires_at': int(time.time()) + 3600}
            payload.update(additional_payload)
            return jwt.encode(payload, key=client_data['key'], algorithm=client_data['algorithm']).decode()
    else:
        raise ValueError(f'client name {client_name} is invalid')



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
from abc import ABCMeta, abstractmethod
from typing import ClassVar, Dict, Optional

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string


def get_provider_cls():
    try:
        SVC_PROVIDER_CLS = settings.PAAS_SERVICE_PROVIDER_CLS
    except AttributeError:
        raise ImproperlyConfigured("PAAS_SERVICE_PROVIDER_CLS is not configured")

    return import_string(SVC_PROVIDER_CLS)


class BaseVendorException(Exception):
    """Base class for vendor exception"""


class ArgumentInvalidError(BaseVendorException):
    """raised when given argument is invalid"""


class OperationFailed(BaseVendorException):
    """raised when operation fails"""


class InstanceData:
    __slots__ = ['credentials', 'config']

    def __init__(self, credentials, config=None):
        self.credentials = credentials
        self.config = config


class BaseProvider(metaclass=ABCMeta):
    SERVICE_NAME: ClassVar[Optional[str]] = None

    @abstractmethod
    def create(self, params: Dict) -> InstanceData:
        raise NotImplementedError

    @abstractmethod
    def delete(self, instanceData: InstanceData):
        raise NotImplementedError

    @abstractmethod
    def patch(self, instanceData: InstanceData, params: Dict) -> InstanceData:
        raise NotImplementedError


class DummyProvider(BaseProvider):
    SERVICE_NAME = 'dummy_svc'

    def create(self, params: Dict) -> InstanceData:
        return InstanceData(credentials={'foo': 'bar'}, config={'foo_config': 'bar_value'})

    def delete(self, instanceData: InstanceData):
        return

    def patch(self, instanceData: InstanceData, params: Dict) -> InstanceData:
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paas_service',
 'paas_service.auth',
 'paas_service.management',
 'paas_service.management.commands',
 'paas_service.migrations']

package_data = \
{'': ['*']}

install_requires = \
['blue-krill>=1.0.15',
 'django-translated-fields',
 'jsonfield',
 'pyjwt>=1.6.4,<2.0.0']

setup_kwargs = {
    'name': 'paas-service',
    'version': '1.1.5',
    'description': 'A Django application for developing BK-PaaS add-on services.',
    'long_description': '# Paas Service\n\n\xe8\x93\x9d\xe9\xb2\xb8 PaaS \xe5\xb9\xb3\xe5\x8f\xb0\xe5\xa2\x9e\xe5\xbc\xba\xe6\x9c\x8d\xe5\x8a\xa1\xe6\xa1\x86\xe6\x9e\xb6\n\n\n## \xe7\x89\x88\xe6\x9c\xac\xe5\x8e\x86\xe5\x8f\xb2\n\n\xe8\xaf\xa6\xe8\xa7\x81 `CHANGES.md`\xe3\x80\x82\n\n## \xe4\xbd\xbf\xe7\x94\xa8\xe6\x8c\x87\xe5\x8d\x97\n\n1. \xe6\x9b\xb4\xe6\x96\xb0 settings\xef\xbc\x9a\n```python\nINSTALLED_APPS = [\n    ...\n    \'paas_service\',\n    ...\n]\n\nMIDDLEWARE = [\n    ...\n    \'paas_service.auth.middleware.VerifiedClientMiddleware\',\n    ...\n]\n\n# \xe6\x95\xb0\xe6\x8d\xae\xe5\xba\x93\xe6\x95\x8f\xe6\x84\x9f\xe5\xad\x97\xe6\xae\xb5\xe5\x8a\xa0\xe5\xaf\x86 Key\nBKKRILL_ENCRYPT_SECRET_KEY = base64.b64encode(b\'\\x01\' * 32)\n\n# \xe4\xb8\x8e PaaS \xe5\xb9\xb3\xe5\x8f\xb0\xe9\x80\x9a\xe4\xbf\xa1\xe7\x9a\x84 JWT \xe4\xbf\xa1\xe6\x81\xaf\nPAAS_SERVICE_JWT_CLIENTS = [\n    {\n        "iss": "paas-v3",\n        "key": "123..........",\n        "algorithm": "HS256",\n    },\n]\n\n# \xe5\xa2\x9e\xe5\xbc\xba\xe6\x9c\x8d\xe5\x8a\xa1\xe4\xbe\x9b\xe5\xba\x94\xe5\x95\x86\xe7\xb1\xbb\nPAAS_SERVICE_PROVIDER_CLS = "svc_xxx.vendor.provider.Provider"\n# \xe5\xa2\x9e\xe5\xbc\xba\xe6\x9c\x8d\xe5\x8a\xa1\xe5\xae\x9e\xe4\xbe\x8b\xe4\xbf\xa1\xe6\x81\xaf\xe6\xb8\xb2\xe6\x9f\x93\xe5\x87\xbd\xe6\x95\xb0\nPAAS_SERVICE_SVC_INSTANCE_RENDER_FUNC = "svc_xxx.vendor.render.render_instance_data"\n\n# \xe8\xae\xbe\xe7\xbd\xae\xe8\xaf\xad\xe8\xa8\x80\xef\xbc\x8c\xe6\xb3\xa8\xe6\x84\x8f\xef\xbc\x9a\xe7\x9b\xae\xe5\x89\x8d\xe5\x9b\xbd\xe9\x99\x85\xe5\x8c\x96\xe5\x8f\xaa\xe6\x94\xaf\xe6\x8c\x81: \xe7\xae\x80\xe4\xbd\x93\xe4\xb8\xad\xe6\x96\x87 \xe5\x92\x8c English\nLANGUAGE_CODE = \'zh-cn\'\n\nLANGUAGES = [("zh-cn", "\xe7\xae\x80\xe4\xbd\x93\xe4\xb8\xad\xe6\x96\x87"), ("en", "English")]\n```\n\n2. \xe5\x8d\x95\xe5\x85\x83\xe6\xb5\x8b\xe8\xaf\x95\n\n\xe9\xa6\x96\xe5\x85\x88\xef\xbc\x8c\xe5\xae\x89\xe8\xa3\x85 pytest\xe3\x80\x81pytest-django\xe3\x80\x82\n\n\xe7\x84\xb6\xe5\x90\x8e\xe6\x89\xa7\xe8\xa1\x8c `make test` \xe8\xbf\x90\xe8\xa1\x8c\xe6\x89\x80\xe6\x9c\x89\xe5\x8d\x95\xe5\x85\x83\xe6\xb5\x8b\xe8\xaf\x95\xe3\x80\x82\n',
    'author': 'blueking',
    'author_email': 'blueking@tencent.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<3.11',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making
蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except
in compliance with the License. You may obtain a copy of the License at

    http://opensource.org/licenses/MIT

Unless required by applicable law or agreed to in writing, software distributed under
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied. See the License for the specific language governing permissions and
limitations under the License.

We undertake not to change the open source license (MIT license) applicable
to the current version of the project delivered to anyone in the future.
"""
import os
import random
import string
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

from django.db import transaction
from paas_service.models import ResourceId, ServiceInstance, ServiceInstanceConfig
from paas_service.serializers import InstanceConfigSLZ

try:
    from django.http.request import limited_parse_qsl as parse_qsl
except ImportError:
    from django.http.request import parse_qsl


def gen_unique_id(
    name: str,
    namespace: str = "default",
    reserve_length: int = 12,
    divide_char: str = "-",
):
    """Generate an unique id via given name
    :param name 原字符串
    :param namespace 字符串命名空间，区分不同领域
    :param reserve_length 原字符串保留长度
    :param divide_char 特殊串和原串的连接符
    """
    prefix = name[:reserve_length]

    with transaction.atomic():
        # create a db instance for getting auto increment id
        resource_id = ResourceId.objects.create(namespace=namespace, uid=name)

        # update uid
        uid = prefix + divide_char + str(resource_id.id)
        resource_id.uid = uid
        resource_id.save(update_fields=["uid"])

    return resource_id.uid


def get_paas_app_info(instance: ServiceInstance) -> Optional[Dict]:
    """Get instance's related PaaS app infomations Dict"""
    config, _ = ServiceInstanceConfig.objects.get_or_create(instance=instance)
    if not config.was_initialized():
        return None

    slz = InstanceConfigSLZ(config)
    return slz.data["paas_app_info"]


def parse_redirect_params(redirect_url: str = None, **kwargs) -> Tuple[str, dict]:
    """
    从 redirect_url 构建出重定向指令的参数
    >>> parse_redirect_params(redirect_url="instance.index?a=1&b=1&b=2&c=3")
    ('instance.index', {'a': '1', 'b': '2', 'c': '3'})
    """

    if not redirect_url:
        return "instance.index", kwargs

    result = urlparse(redirect_url)
    params = dict(parse_qsl(qs=result.query))
    params.update(kwargs)
    return result.path, params


def generate_password(length=10):
    """生成随机的由大小写字母和数字组成的且至少包含 1 位 数字的密码."""
    password_chars = [random.choice(string.ascii_letters + string.digits) for _ in range(length)]
    password_chars.append(random.choice(string.digits))
    random.shuffle(password_chars)
    return "".join(password_chars)


# WR Algorithm start


class WRItem:
    """A single weighted-random items"""

    @classmethod
    def from_dict(cls, d):
        return WRItem(values=d["values"], weight=d.get("weight", 0))

    def __init__(self, values, weight=0):
        self.values = values
        self.weight = weight

    def __str__(self):
        return f"values={self.values} weight={self.weight}"


class WRItemList:
    # if the precision was set to 10, then an item which's weight is below 10% of the total weight
    # of all items is considered as zero and will never be choosed.
    precision = 100

    @classmethod
    def from_json(cls, items_list: List[Dict]):
        """Generate a list object from json, an example items_list:

        [
            {"values": ANY_THING, "weight": 10},
            {"values": ANY_THING, "weight": 3},
        ]
        """
        items = []
        for data in items_list:
            items.append(WRItem.from_dict(data))
        return WRItemList(items)

    def __init__(self, items: List[WRItem]):
        self.items = items
        self.initalize()

    def initalize(self):
        """Caculate weight and generate a list of size 100 in order to archive the random choice"""
        total_weight = sum(item.weight for item in self.items)
        if not total_weight:
            raise ValueError("no valid items given")

        self._list_of_choices = []
        for item in self.items:
            repeats = int(self.precision * (item.weight / total_weight))
            self._list_of_choices += [item] * repeats

    def get(self) -> WRItem:
        """Pick a item based on weighted-random algorithm"""
        return random.choice(self._list_of_choices)


# WR Algorithm end


def get_node_ip():
    """获取当前节点IP. 使用 helm 部署时, 在 values.yaml 中注入 NODE_IP 这个环境变量."""
    return os.environ.get("NODE_IP", None)


class Base36Handler:
    # keep lowercase
    BASE36 = "0123456789abcdefghijklmnopqrstuvwxyz"

    @classmethod
    def encode(cls, num: int, alphabet=BASE36):
        """Encode a positive number in Base X

        Arguments:
        - `num`: The number to encode
        - `alphabet`: The alphabet to use for encoding
        """
        if num == 0:
            return alphabet[0]
        arr = []
        base = len(alphabet)
        while num:
            num, rem = divmod(num, base)
            arr.append(alphabet[rem])
        arr.reverse()
        return ''.join(arr)

    @classmethod
    def decode(cls, encoded: str, alphabet=BASE36):
        """Decode a Base X encoded string into the number

        Arguments:
        - `string`: The encoded string
        - `alphabet`: The alphabet to use for encoding
        """
        base = len(alphabet)
        str_len = len(encoded)
        num = 0

        idx = 0
        for char in encoded:
            power = str_len - (idx + 1)
            num += alphabet.index(char) * (base ** power)
            idx += 1

        return num
